@@ -12,6 +12,7 @@
 #include "averr.h"
 #include <gbm.h>
 
+#define HAVE_CH_LAYOUT (LIBAVUTIL_VERSION_INT >= AV_VERSION_INT(57, 28, 100))
 
 static const AVRational US_RATIONAL{1,1000000} ;
 
@@ -79,7 +80,7 @@ void FrameWriter::load_codec_options(AVDictionary **dict)
 
     for (auto& opt : params.codec_options)
     {
-        std::cout << "Setting codec option: " << opt.first << "=" << opt.second << std::endl;
+        std::cerr << "Setting codec option: " << opt.first << "=" << opt.second << std::endl;
         av_dict_set(dict, opt.first.c_str(), opt.second.c_str(), 0);
     }
 }
@@ -88,7 +89,7 @@ void FrameWriter::load_audio_codec_options(AVDictionary **dict)
 {
     for (auto& opt : params.audio_codec_options)
     {
-        std::cout << "Setting codec option: " << opt.first << "=" << opt.second << std::endl;
+        std::cerr << "Setting codec option: " << opt.first << "=" << opt.second << std::endl;
         av_dict_set(dict, opt.first.c_str(), opt.second.c_str(), 0);
     }
 }
@@ -163,7 +164,7 @@ static AVPixelFormat get_drm_av_format(int fmt)
             return drm_av_format_table[i].av;
         }
     }
-    std::cerr << "Failed to find AV format for" << fmt;
+    std::cerr << "Failed to find AV format for" << fmt << std::endl;
     return AV_PIX_FMT_RGBA;
 }
 
@@ -180,6 +181,10 @@ AVPixelFormat FrameWriter::lookup_pixel_format(std::string pix_fmt)
 
 AVPixelFormat FrameWriter::handle_buffersink_pix_fmt(const AVCodec *codec)
 {
+    /* If using the default codec and no pixel format is specified,
+     * set the format to yuv420p for web friendly output by default */
+    if (params.codec == DEFAULT_CODEC && params.pix_fmt.empty())
+        params.pix_fmt = "yuv420p";
 
     // Return with user chosen format
     if (!params.pix_fmt.empty())
@@ -230,25 +235,13 @@ void FrameWriter::init_video_filters(const AVCodec *codec)
     }
 
     if (this->hw_device_context) {
-        this->hw_frame_context = av_hwframe_ctx_alloc(this->hw_device_context);
-        AVHWFramesContext *hwfc = reinterpret_cast<AVHWFramesContext*>(this->hw_frame_context->data);
-        hwfc->format = AV_PIX_FMT_VAAPI;
-        hwfc->sw_format = AV_PIX_FMT_NV12;
-        hwfc->width = params.width;
-        hwfc->height = params.height;
-        int err = av_hwframe_ctx_init(this->hw_frame_context);
-        if (err < 0) {
-            std::cerr << "Cannot create hw frames context: " << averr(err) << std::endl;
-            exit(-1);
-        }
-
         this->hw_frame_context_in = av_hwframe_ctx_alloc(this->hw_device_context);
-        hwfc = reinterpret_cast<AVHWFramesContext*>(this->hw_frame_context_in->data);
+        AVHWFramesContext *hwfc = reinterpret_cast<AVHWFramesContext*>(this->hw_frame_context_in->data);
         hwfc->format = AV_PIX_FMT_VAAPI;
         hwfc->sw_format = get_drm_av_format(params.drm_format);
         hwfc->width = params.width;
         hwfc->height = params.height;
-        err = av_hwframe_ctx_init(this->hw_frame_context_in);
+        int err = av_hwframe_ctx_init(this->hw_frame_context_in);
         if (err < 0) {
             std::cerr << "Cannot create hw frames context: " << averr(err) << std::endl;
             exit(-1);
@@ -338,7 +331,7 @@ void FrameWriter::init_video_filters(const AVCodec *codec)
         exit(-1);
     }
 
-    std::cout << "Using video filter: " << params.video_filter << std::endl;
+    std::cerr << "Using video filter: " << params.video_filter << std::endl;
 
     err = avfilter_graph_parse_ptr(this->videoFilterGraph,
         params.video_filter.c_str(), &inputs, &outputs, NULL);
@@ -365,9 +358,9 @@ void FrameWriter::init_video_filters(const AVCodec *codec)
     }
 
     if (params.enable_ffmpeg_debug_output) {
-        std::cout << std::string(80,'#') << std::endl ;
-        std::cout << avfilter_graph_dump(this->videoFilterGraph,0) << "\n";
-        std::cout << std::string(80,'#') << std::endl ;
+        std::cerr << std::string(80,'#') << std::endl ;
+        std::cerr << avfilter_graph_dump(this->videoFilterGraph,0) << "\n";
+        std::cerr << std::string(80,'#') << std::endl ;
     }
 
 
@@ -381,6 +374,8 @@ void FrameWriter::init_video_filters(const AVCodec *codec)
     this->videoCodecCtx->framerate = filter_output->frame_rate; // can be 1/0 if unknown
     this->videoCodecCtx->sample_aspect_ratio = filter_output->sample_aspect_ratio;
 
+    this->hw_frame_context = av_buffersink_get_hw_frames_ctx(
+        this->videoFilterSinkCtx);
     avfilter_inout_free(&inputs);
     avfilter_inout_free(&outputs);
 }
@@ -410,7 +405,7 @@ void FrameWriter::init_video_stream()
     videoCodecCtx->time_base  = US_RATIONAL;
     videoCodecCtx->color_range = AVCOL_RANGE_JPEG;
     if (params.framerate) {
-    std::cout << "Framerate: " << params.framerate << std::endl;
+    std::cerr << "Framerate: " << params.framerate << std::endl;
     }
 
     if (params.bframes != -1)
@@ -451,7 +446,23 @@ void FrameWriter::init_video_stream()
     }
 }
 
-#ifdef HAVE_PULSE
+#ifdef HAVE_AUDIO
+#if HAVE_CH_LAYOUT
+static uint64_t get_codec_channel_layout(const AVCodec *codec)
+{
+    int i = 0;
+    if (!codec->ch_layouts)
+        return AV_CH_LAYOUT_STEREO;
+    while (1) {
+        if (!av_channel_layout_check(&codec->ch_layouts[i]))
+            break;
+        if (codec->ch_layouts[i].u.mask == AV_CH_LAYOUT_STEREO)
+            return codec->ch_layouts[i].u.mask;
+        i++;
+    }
+    return codec->ch_layouts[0].u.mask;
+}
+#else
 static uint64_t get_codec_channel_layout(const AVCodec *codec)
 {
       int i = 0;
@@ -466,6 +477,7 @@ static uint64_t get_codec_channel_layout(const AVCodec *codec)
       }
       return codec->channel_layouts[0];
 }
+#endif
 
 static enum AVSampleFormat get_codec_auto_sample_fmt(const AVCodec *codec)
 {
@@ -498,16 +510,16 @@ static enum AVSampleFormat convert_codec_sample_fmt(const AVCodec *codec, std::s
     static enum AVSampleFormat converted_fmt = av_get_sample_fmt(requested_fmt.c_str());
     if (converted_fmt == AV_SAMPLE_FMT_NONE)
     {
-	std::cout << "Failed to find the given sample format: " << requested_fmt << std::endl;
-	std::exit(-1);
+        std::cerr << "Failed to find the given sample format: " << requested_fmt << std::endl;
+        std::exit(-1);
     } else if (!codec->sample_fmts || check_fmt_available(codec, converted_fmt))
     {
-        std::cout << "Using sample format " << av_get_sample_fmt_name(converted_fmt) << " for audio codec " << codec->name << std::endl;
+        std::cerr << "Using sample format " << av_get_sample_fmt_name(converted_fmt) << " for audio codec " << codec->name << std::endl;
         return converted_fmt;
     } else
     {
-	std::cout << "Codec " << codec->name << " does not support sample format " << av_get_sample_fmt_name(converted_fmt) << std::endl;
-	std::exit(-1);
+        std::cerr << "Codec " << codec->name << " does not support sample format " << av_get_sample_fmt_name(converted_fmt) << std::endl;
+        std::exit(-1);
     }
 }
 
@@ -534,15 +546,19 @@ void FrameWriter::init_audio_stream()
     if (params.sample_fmt.size() == 0) 
     {
         audioCodecCtx->sample_fmt = get_codec_auto_sample_fmt(codec);
-        std::cout << "Choosing sample format " << av_get_sample_fmt_name(audioCodecCtx->sample_fmt) << " for audio codec " << codec->name << std::endl;
+        std::cerr << "Choosing sample format " << av_get_sample_fmt_name(audioCodecCtx->sample_fmt) << " for audio codec " << codec->name << std::endl;
     } else 
     {
         audioCodecCtx->sample_fmt = convert_codec_sample_fmt(codec, params.sample_fmt);
     }
+#if HAVE_CH_LAYOUT
+    av_channel_layout_from_mask(&audioCodecCtx->ch_layout, get_codec_channel_layout(codec));
+#else
     audioCodecCtx->channel_layout = get_codec_channel_layout(codec);
+    audioCodecCtx->channels = av_get_channel_layout_nb_channels(audioCodecCtx->channel_layout);
+#endif
     audioCodecCtx->sample_rate = params.sample_rate;
     audioCodecCtx->time_base = (AVRational) { 1, 1000 };
-    audioCodecCtx->channels = av_get_channel_layout_nb_channels(audioCodecCtx->channel_layout);
 
     if (fmtCtx->oformat->flags & AVFMT_GLOBALHEADER)
         audioCodecCtx->flags |= AV_CODEC_FLAG_GLOBAL_HEADER;
@@ -565,8 +581,14 @@ void FrameWriter::init_audio_stream()
     av_opt_set_int(swrCtx, "out_sample_rate", audioCodecCtx->sample_rate, 0);
     av_opt_set_sample_fmt(swrCtx, "in_sample_fmt", AV_SAMPLE_FMT_FLT, 0);
     av_opt_set_sample_fmt(swrCtx, "out_sample_fmt", audioCodecCtx->sample_fmt, 0);
+#if HAVE_CH_LAYOUT
+    AVChannelLayout in_chlayout = AV_CHANNEL_LAYOUT_STEREO;
+    av_opt_set_chlayout(swrCtx, "in_chlayout", &in_chlayout, 0);
+    av_opt_set_chlayout(swrCtx, "out_chlayout", &audioCodecCtx->ch_layout, 0);
+#else
     av_opt_set_channel_layout(swrCtx, "in_channel_layout", AV_CH_LAYOUT_STEREO, 0);
     av_opt_set_channel_layout(swrCtx, "out_channel_layout", audioCodecCtx->channel_layout, 0);
+#endif
 
     if (swr_init(swrCtx))
     {
@@ -586,7 +608,7 @@ void FrameWriter::init_audio_stream()
 void FrameWriter::init_codecs()
 {
     init_video_stream();
-#ifdef HAVE_PULSE
+#ifdef HAVE_AUDIO
     if (params.enable_audio)
         init_audio_stream();
 #endif
@@ -817,7 +839,7 @@ bool FrameWriter::add_frame(struct gbm_bo *bo, int64_t usec, bool y_invert)
     return push_frame(frame, usec);
 }
 
-#ifdef HAVE_PULSE
+#ifdef HAVE_AUDIO
 #define SRC_RATE 1e6
 #define DST_RATE 1e3
 
@@ -856,7 +878,11 @@ void FrameWriter::add_audio(const void* buffer)
     AVFrame *inputf = av_frame_alloc();
     inputf->sample_rate    = params.sample_rate;
     inputf->format         = AV_SAMPLE_FMT_FLT;
+#if HAVE_CH_LAYOUT
+    inputf->ch_layout = (AVChannelLayout) AV_CHANNEL_LAYOUT_STEREO;
+#else
     inputf->channel_layout = AV_CH_LAYOUT_STEREO;
+#endif
     inputf->nb_samples     = audioCodecCtx->frame_size;
 
     av_frame_get_buffer(inputf, 0);
@@ -865,7 +891,11 @@ void FrameWriter::add_audio(const void* buffer)
     AVFrame *outputf = av_frame_alloc();
     outputf->format         = audioCodecCtx->sample_fmt;
     outputf->sample_rate    = audioCodecCtx->sample_rate;
+#if HAVE_CH_LAYOUT
+    av_channel_layout_copy(&outputf->ch_layout, &audioCodecCtx->ch_layout);
+#else
     outputf->channel_layout = audioCodecCtx->channel_layout;
+#endif
     outputf->nb_samples     = audioCodecCtx->frame_size;
     av_frame_get_buffer(outputf, 0);
 
@@ -888,7 +918,7 @@ void FrameWriter::finish_frame(AVCodecContext *enc_ctx, AVPacket& pkt)
         av_packet_rescale_ts(&pkt, videoCodecCtx->time_base, videoStream->time_base);
         pkt.stream_index = videoStream->index;
     }
-#ifdef HAVE_PULSE
+#ifdef HAVE_AUDIO
     else
     {
         av_packet_rescale_ts(&pkt, (AVRational){ 1, 1000 }, audioStream->time_base);
@@ -909,7 +939,7 @@ void FrameWriter::finish_frame(AVCodecContext *enc_ctx, AVPacket& pkt)
         params.write_aborted_flag = true;
     }
     av_packet_unref(&pkt);
-#ifdef HAVE_PULSE
+#ifdef HAVE_AUDIO
     if (params.enable_audio)
         fmt_mutex.unlock();
 #endif
@@ -921,7 +951,7 @@ FrameWriter::~FrameWriter()
     AVPacket *pkt = av_packet_alloc();
 
     encode(videoCodecCtx, NULL, pkt);
-#ifdef HAVE_PULSE
+#ifdef HAVE_AUDIO
     if (params.enable_audio)
     {
         encode(audioCodecCtx, NULL, pkt);
@@ -936,7 +966,7 @@ FrameWriter::~FrameWriter()
 
     // Freeing all the allocated memory:
     avcodec_free_context(&videoCodecCtx);
-#ifdef HAVE_PULSE
+#ifdef HAVE_AUDIO
     if (params.enable_audio)
         avcodec_free_context(&audioCodecCtx);
 #endif
