@@ -1,6 +1,7 @@
 #define _XOPEN_SOURCE 700
 #define _POSIX_C_SOURCE 199309L
 #include <iostream>
+#include <optional>
 
 #include <list>
 #include <string>
@@ -9,7 +10,6 @@
 #include <atomic>
 #include <getopt.h>
 
-#include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -20,19 +20,19 @@
 #include <wayland-client-protocol.h>
 #include <gbm.h>
 #include <fcntl.h>
+#include <xf86drm.h>
 
 #include "frame-writer.hpp"
 #include "buffer-pool.hpp"
 #include "wlr-screencopy-unstable-v1-client-protocol.h"
 #include "xdg-output-unstable-v1-client-protocol.h"
 #include "linux-dmabuf-unstable-v1-client-protocol.h"
-#include "wl-drm-client-protocol.h"
 
 #include "config.h"
 
-#ifdef HAVE_PULSE
-#include "pulse.hpp"
-PulseReaderParams pulseParams;
+#ifdef HAVE_AUDIO
+#include "audio.hpp"
+AudioReaderParams audioParams;
 #endif
 
 #define MAX_FRAME_FAILURES 16
@@ -50,7 +50,6 @@ static struct wl_shm *shm = NULL;
 static struct zxdg_output_manager_v1 *xdg_output_manager = NULL;
 static struct zwlr_screencopy_manager_v1 *screencopy_manager = NULL;
 static struct zwp_linux_dmabuf_v1 *dmabuf = NULL;
-static struct wl_drm *drm = NULL;
 void request_next_frame();
 
 struct wf_recorder_output
@@ -123,8 +122,10 @@ const zxdg_output_v1_listener xdg_output_implementation = {
 struct wf_buffer : public buffer_pool_buf
 {
     struct gbm_bo *bo = nullptr;
+    zwp_linux_buffer_params_v1 *params = nullptr;
     struct wl_buffer *wl_buffer = nullptr;
     void *data = nullptr;
+    size_t size = 0;
     enum wl_shm_format format;
     int drm_format;
     int width, height, stride;
@@ -189,10 +190,33 @@ static struct wl_buffer *create_shm_buffer(uint32_t fmt,
     return buffer;
 }
 
+void free_shm_buffer(wf_buffer& buffer)
+{
+    if (buffer.wl_buffer == NULL)
+    {
+        return;
+    }
+
+    munmap(buffer.data, buffer.size);
+    wl_buffer_destroy(buffer.wl_buffer);
+    buffer.wl_buffer = NULL;
+}
+
 static bool use_damage = true;
 static bool use_dmabuf = false;
 static bool use_hwupload = false;
 
+static uint32_t wl_shm_to_drm_format(uint32_t format)
+{
+    if (format == WL_SHM_FORMAT_ARGB8888) {
+        return GBM_FORMAT_ARGB8888;
+    } else if (format == WL_SHM_FORMAT_XRGB8888) {
+        return GBM_FORMAT_XRGB8888;
+    } else {
+        return format;
+    }
+}
+
 static void frame_handle_buffer(void *, struct zwlr_screencopy_frame_v1 *frame, uint32_t format,
     uint32_t width, uint32_t height, uint32_t stride)
 {
@@ -201,8 +225,9 @@ static void frame_handle_buffer(void *, struct zwlr_screencopy_frame_v1 *frame,
     }
 
     auto& buffer = buffers.capture();
-
+    auto old_format = buffer.format;
     buffer.format = (wl_shm_format)format;
+    buffer.drm_format = wl_shm_to_drm_format(format);
     buffer.width = width;
     buffer.height = height;
     buffer.stride = stride;
@@ -213,7 +238,8 @@ static void frame_handle_buffer(void *, struct zwlr_screencopy_frame_v1 *frame,
     if (buffer.height % 2)
         buffer.height -= 1;
 
-    if (!buffer.wl_buffer) {
+    if (!buffer.wl_buffer || old_format != format) {
+        free_shm_buffer(buffer);
         buffer.wl_buffer =
             create_shm_buffer(format, width, height, stride, &buffer.data);
     }
@@ -307,12 +333,22 @@ static void frame_handle_linux_dmabuf(void *, struct zwlr_screencopy_frame_v1 *f
 
     auto& buffer = buffers.capture();
 
+    auto old_format = buffer.format;
     buffer.format = drm_to_wl_shm_format(format);
     buffer.drm_format = format;
     buffer.width = width;
     buffer.height = height;
 
-    if (!buffer.wl_buffer) {
+    if (!buffer.wl_buffer || (old_format != buffer.format)) {
+        if (buffer.bo) {
+            if (buffer.wl_buffer) {
+                wl_buffer_destroy(buffer.wl_buffer);
+            }
+
+            zwp_linux_buffer_params_v1_destroy(buffer.params);
+            gbm_bo_destroy(buffer.bo);
+        }
+
         const uint64_t modifier = 0; // DRM_FORMAT_MOD_LINEAR
         buffer.bo = gbm_bo_create_with_modifiers(gbm_device, buffer.width,
             buffer.height, format, &modifier, 1);
@@ -330,19 +366,17 @@ static void frame_handle_linux_dmabuf(void *, struct zwlr_screencopy_frame_v1 *f
 
         buffer.stride = gbm_bo_get_stride(buffer.bo);
 
-        struct zwp_linux_buffer_params_v1 *params =
-            zwp_linux_dmabuf_v1_create_params(dmabuf);
+        buffer.params = zwp_linux_dmabuf_v1_create_params(dmabuf);
 
         uint64_t mod = gbm_bo_get_modifier(buffer.bo);
-        zwp_linux_buffer_params_v1_add(params,
+        zwp_linux_buffer_params_v1_add(buffer.params,
             gbm_bo_get_fd(buffer.bo), 0,
             gbm_bo_get_offset(buffer.bo, 0),
             gbm_bo_get_stride(buffer.bo),
             mod >> 32, mod & 0xffffffff);
 
-        zwp_linux_buffer_params_v1_add_listener(params, &params_listener, frame);
-
-        zwp_linux_buffer_params_v1_create(params, buffer.width,
+        zwp_linux_buffer_params_v1_add_listener(buffer.params, &params_listener, frame);
+        zwp_linux_buffer_params_v1_create(buffer.params, buffer.width,
             buffer.height, format, 0);
     } else {
         if (use_damage) {
@@ -366,24 +400,65 @@ static const struct zwlr_screencopy_frame_v1_listener frame_listener = {
     .buffer_done = frame_handle_buffer_done,
 };
 
-static void drm_handle_device(void *, struct wl_drm *, const char *name) {
-    drm_device_name = name;
+static void dmabuf_feedback_done(void *, struct zwp_linux_dmabuf_feedback_v1 *feedback)
+{
+    zwp_linux_dmabuf_feedback_v1_destroy(feedback);
 }
 
-static void drm_handle_format(void *, struct wl_drm *, uint32_t) {
+static void dmabuf_feedback_format_table(void *, struct zwp_linux_dmabuf_feedback_v1 *,
+    int32_t fd, uint32_t)
+{
+    close(fd);
 }
 
-static void drm_handle_authenticated(void *, struct wl_drm *) {
+static void dmabuf_feedback_main_device(void *, struct zwp_linux_dmabuf_feedback_v1 *,
+    struct wl_array *device)
+{
+    dev_t dev_id;
+    memcpy(&dev_id, device->data, device->size);
+
+    drmDevice *dev = NULL;
+    if (drmGetDeviceFromDevId(dev_id, 0, &dev) != 0) {
+        std::cerr << "Failed to get DRM device from dev id " << strerror(errno) << std::endl;
+        return;
+    }
+
+    if (dev->available_nodes & (1 << DRM_NODE_RENDER)) {
+        drm_device_name = dev->nodes[DRM_NODE_RENDER];
+    } else if (dev->available_nodes & (1 << DRM_NODE_PRIMARY)) {
+        drm_device_name = dev->nodes[DRM_NODE_PRIMARY];
+    }
+
+    drmFreeDevice(&dev);
 }
 
-static void drm_handle_capabilities(void *, struct wl_drm *, uint32_t) {
+static void dmabuf_feedback_tranche_done(void *, struct zwp_linux_dmabuf_feedback_v1 *)
+{
 }
 
-static const struct wl_drm_listener drm_listener = {
-    .device = drm_handle_device,
-    .format = drm_handle_format,
-    .authenticated = drm_handle_authenticated,
-    .capabilities = drm_handle_capabilities,
+static void dmabuf_feedback_tranche_target_device(void *, struct zwp_linux_dmabuf_feedback_v1 *,
+    struct wl_array *)
+{
+}
+
+static void dmabuf_feedback_tranche_formats(void *, struct zwp_linux_dmabuf_feedback_v1 *,
+    struct wl_array *)
+{
+}
+
+static void dmabuf_feedback_tranche_flags(void *, struct zwp_linux_dmabuf_feedback_v1 *,
+    uint32_t)
+{
+}
+
+static const struct zwp_linux_dmabuf_feedback_v1_listener dmabuf_feedback_listener = {
+    .done = dmabuf_feedback_done,
+    .format_table = dmabuf_feedback_format_table,
+    .main_device = dmabuf_feedback_main_device,
+    .tranche_done = dmabuf_feedback_tranche_done,
+    .tranche_target_device = dmabuf_feedback_tranche_target_device,
+    .tranche_formats = dmabuf_feedback_tranche_formats,
+    .tranche_flags = dmabuf_feedback_tranche_flags,
 };
 
 static void handle_global(void*, struct wl_registry *registry,
@@ -413,12 +488,12 @@ static void handle_global(void*, struct wl_registry *registry,
     else if (strcmp(interface, zwp_linux_dmabuf_v1_interface.name) == 0)
     {
         dmabuf = (zwp_linux_dmabuf_v1*) wl_registry_bind(registry, name,
-            &zwp_linux_dmabuf_v1_interface, 3);
-    }
-    else if (strcmp(interface, wl_drm_interface.name) == 0)
-    {
-        drm = (wl_drm*) wl_registry_bind(registry, name, &wl_drm_interface, 1);
-        wl_drm_add_listener(drm, &drm_listener, NULL);
+            &zwp_linux_dmabuf_v1_interface, 4);
+        if (dmabuf) {
+            struct zwp_linux_dmabuf_feedback_v1 *feedback =
+                zwp_linux_dmabuf_v1_get_default_feedback(dmabuf);
+            zwp_linux_dmabuf_feedback_v1_add_listener(feedback, &dmabuf_feedback_listener, NULL);
+        }
     }
 }
 
@@ -486,10 +561,12 @@ static void write_loop(FrameWriterParams params)
     }
     pthread_sigmask(SIG_BLOCK, &sigset, NULL);
 
-#ifdef HAVE_PULSE
-    std::unique_ptr<PulseReader> pr;
+#ifdef HAVE_AUDIO
+    std::unique_ptr<AudioReader> pr;
 #endif
 
+    std::optional<uint64_t> first_frame_ts;
+
     while(!exit_main_loop)
     {
         // wait for frame to become available
@@ -516,39 +593,62 @@ static void write_loop(FrameWriterParams params)
             params.stride = buffer.stride;
             frame_writer = std::unique_ptr<FrameWriter> (new FrameWriter(params));
 
-#ifdef HAVE_PULSE
+#ifdef HAVE_AUDIO
             if (params.enable_audio)
             {
-                pulseParams.audio_frame_size = frame_writer->get_audio_buffer_size();
-                pulseParams.sample_rate = params.sample_rate;
-                pr = std::unique_ptr<PulseReader> (new PulseReader(pulseParams));
-                pr->start();
+                audioParams.audio_frame_size = frame_writer->get_audio_buffer_size();
+                audioParams.sample_rate = params.sample_rate;
+                pr = std::unique_ptr<AudioReader> (AudioReader::create(audioParams));
+                if (pr)
+                {
+                    pr->start();
+                }
             }
 #endif
         }
 
-        bool do_cont = false;
-
-        if (use_dmabuf) {
-            if (use_hwupload) {
-                uint32_t stride = 0;
-                void *map_data = NULL;
-                void *data = gbm_bo_map(buffer.bo, 0, 0, buffer.width, buffer.height,
-                    GBM_BO_TRANSFER_READ, &stride, &map_data);
-                if (!data) {
-                    std::cerr << "Failed to map bo" << std::endl;
-                    break;
-                }
-                do_cont = frame_writer->add_frame((unsigned char*)data,
-                    buffer.base_usec, buffer.y_invert);
-                gbm_bo_unmap(buffer.bo, map_data);
+        bool drop = false;
+        uint64_t sync_timestamp = 0;
+        if (first_frame_ts.has_value()) {
+            sync_timestamp = buffer.base_usec - first_frame_ts.value();
+        } else if (pr) {
+            if (!pr->get_time_base() || pr->get_time_base() > buffer.base_usec) {
+                drop = true;
             } else {
-                do_cont = frame_writer->add_frame(buffer.bo,
-                    buffer.base_usec, buffer.y_invert);
+                first_frame_ts = pr->get_time_base();
+                sync_timestamp = buffer.base_usec - first_frame_ts.value();
             }
         } else {
-            do_cont = frame_writer->add_frame((unsigned char*)buffer.data,
-                buffer.base_usec, buffer.y_invert);
+            sync_timestamp = 0;
+            first_frame_ts = buffer.base_usec;
+        }
+
+        bool do_cont = false;
+
+        if (!drop) {
+            if (use_dmabuf) {
+                if (use_hwupload) {
+                    uint32_t stride = 0;
+                    void *map_data = NULL;
+                    void *data = gbm_bo_map(buffer.bo, 0, 0, buffer.width, buffer.height,
+                        GBM_BO_TRANSFER_READ, &stride, &map_data);
+                    if (!data) {
+                        std::cerr << "Failed to map bo" << std::endl;
+                        break;
+                    }
+                    do_cont = frame_writer->add_frame((unsigned char*)data,
+                        sync_timestamp, buffer.y_invert);
+                    gbm_bo_unmap(buffer.bo, map_data);
+                } else {
+                    do_cont = frame_writer->add_frame(buffer.bo,
+                        sync_timestamp, buffer.y_invert);
+                }
+            } else {
+                do_cont = frame_writer->add_frame((unsigned char*)buffer.data,
+                    sync_timestamp, buffer.y_invert);
+            }
+        } else {
+            do_cont = true;
         }
 
         frame_writer_mutex.unlock();
@@ -561,9 +661,9 @@ static void write_loop(FrameWriterParams params)
     }
 
     std::lock_guard<std::mutex> lock(frame_writer_mutex);
-    /* Free the PulseReader connection first. This way it'd flush any remaining
+    /* Free the AudioReader connection first. This way it'd flush any remaining
      * frames to the FrameWriter */
-#ifdef HAVE_PULSE
+#ifdef HAVE_AUDIO
     pr = nullptr;
 #endif
     frame_writer = nullptr;
@@ -580,11 +680,11 @@ static bool user_specified_overwrite(std::string filename)
     if (stat (filename.c_str(), &buffer) == 0 && !S_ISCHR(buffer.st_mode))
     {
         std::string input;
-        std::cout << "Output file \"" << filename << "\" exists. Overwrite? Y/n: ";
+        std::cerr << "Output file \"" << filename << "\" exists. Overwrite? Y/n: ";
         std::getline(std::cin, input);
         if (input.size() && input[0] != 'Y' && input[0] != 'y')
         {
-            std::cout << "Use -f to specify the file name." << std::endl;
+            std::cerr << "Use -f to specify the file name." << std::endl;
             return false;
 	}
     }
@@ -709,7 +809,7 @@ static wf_recorder_output* detect_output_from_region(const capture_region& regio
         const capture_region output_region{wo.x, wo.y, wo.width, wo.height};
         if (region.contained_in(output_region))
         {
-            std::cout << "Detected output based on geometry: " << wo.name << std::endl;
+            std::cerr << "Detected output based on geometry: " << wo.name << std::endl;
             return &wo;
         }
     }
@@ -726,12 +826,12 @@ Screen recording of wlroots-based compositors
 With no FILE, start recording the current screen.
 
 Use Ctrl+C to stop.)");
-#ifdef HAVE_PULSE
+#ifdef HAVE_AUDIO
     printf(R"(
 
   -a, --audio[=DEVICE]      Starts recording the screen with audio.
                             [=DEVICE] argument is optional.
-                            In case you want to specify the pulseaudio device which will capture
+                            In case you want to specify the audio device which will capture
                             the audio, you can run this command with the name of that device.
                             You can find your device by running: pactl list sources | grep Name
                             Specify device like this: -a<device> or --audio=<device>)");
@@ -792,6 +892,9 @@ Use Ctrl+C to stop.)");
     
   -B. --buffrate            This option is used to specify the buffers expected framerate. this 
                             may help when encoders are expecting specific or limited framerate.
+
+  --audio-backend           Specifies the audio backend among the available backends, for ex.
+                            --audio-backend=pipewire
   
   -C, --audio-codec         Specifies the codec of the audio. These can be found by running:
                             ffmpeg -encoders
@@ -804,9 +907,11 @@ Use Ctrl+C to stop.)");
   
   -P, --audio-codec-param   Change the audio codec parameters.
                             -P <option_name>=<option_value>
+  
+  -y, --overwrite           Force overwriting the output file without prompting.
 
 Examples:)");
-#ifdef HAVE_PULSE
+#ifdef HAVE_AUDIO
     printf(R"(
 
   Video Only:)");
@@ -820,7 +925,7 @@ Examples:)");
   - wf-recorder -f <filename>.ext       Records the video. Use Ctrl+C to stop recording.
                                         The video file will be stored as <filename>.ext in the
                                         current working directory.)");
-#ifdef HAVE_PULSE
+#ifdef HAVE_AUDIO
     printf(R"(
 
   Video and Audio:
@@ -887,6 +992,7 @@ int main(int argc, char *argv[])
     FrameWriterParams params = FrameWriterParams(exit_main_loop);
     params.file = "recording." + std::string(DEFAULT_CONTAINER_FORMAT);
     params.codec = DEFAULT_CODEC;
+    params.pix_fmt = DEFAULT_PIX_FMT;
     params.audio_codec = DEFAULT_AUDIO_CODEC;
     params.sample_rate = DEFAULT_AUDIO_SAMPLE_RATE;
     params.enable_ffmpeg_debug_output = false;
@@ -896,6 +1002,7 @@ int main(int argc, char *argv[])
     constexpr const char* default_cmdline_output = "interactive";
     std::string cmdline_output = default_cmdline_output;
     bool force_no_dmabuf = false;
+    bool force_overwrite = false;
 
     struct option opts[] = {
         { "output",            required_argument, NULL, 'o' },
@@ -906,6 +1013,7 @@ int main(int argc, char *argv[])
         { "codec-param",       required_argument, NULL, 'p' },
         { "framerate",         required_argument, NULL, 'r' },
         { "pixel-format",      required_argument, NULL, 'x' },
+        { "audio-backend",     required_argument, NULL, '*' },
         { "audio-codec",       required_argument, NULL, 'C' },
         { "audio-codec-param", required_argument, NULL, 'P' },
         { "sample-rate",       required_argument, NULL, 'R' },
@@ -920,11 +1028,12 @@ int main(int argc, char *argv[])
         { "buffrate",          required_argument, NULL, 'B' },
         { "version",           no_argument,       NULL, 'v' },
         { "no-damage",         no_argument,       NULL, 'D' },
+        { "overwrite",         no_argument,       NULL, 'y' },
         { 0,                   0,                 NULL,  0  }
     };
 
     int c, i;
-    while((c = getopt_long(argc, argv, "o:f:m:g:c:p:r:x:C:P:R:X:d:b:B:la::hvDF:", opts, &i)) != -1)
+    while((c = getopt_long(argc, argv, "o:f:m:g:c:p:r:x:C:P:R:X:d:b:B:la::hvDF:y", opts, &i)) != -1)
     {
         switch(c)
         {
@@ -989,11 +1098,11 @@ int main(int argc, char *argv[])
                 break;
 
             case 'a':
-#ifdef HAVE_PULSE
+#ifdef HAVE_AUDIO
                 params.enable_audio = true;
-                pulseParams.audio_source = optarg ? strdup(optarg) : NULL;
+                audioParams.audio_source = optarg ? strdup(optarg) : NULL;
 #else
-                std::cerr << "Cannot record audio. Built without pulse support." << std::endl;
+                std::cerr << "Cannot record audio. Built without audio support." << std::endl;
                 return EXIT_FAILURE;
 #endif
                 break;
@@ -1022,12 +1131,20 @@ int main(int argc, char *argv[])
                 force_no_dmabuf = true;
                 break;
 
+            case 'y':
+                force_overwrite = true;
+                break;
+
+            case '*':
+                audioParams.audio_backend = optarg;
+                break;
+
             default:
                 printf("Unsupported command line argument %s\n", optarg);
         }
     }
 
-    if (!user_specified_overwrite(params.file))
+    if (!force_overwrite && !user_specified_overwrite(params.file))
     {
         return EXIT_FAILURE;
     }
@@ -1045,7 +1162,7 @@ int main(int argc, char *argv[])
 
     if (params.codec.find("vaapi") != std::string::npos)
     {
-        std::cout << "using VA-API, trying to enable DMA-BUF capture..." << std::endl;
+        std::cerr << "using VA-API, trying to enable DMA-BUF capture..." << std::endl;
 
         // try compositor device if not explicitly set
         if (params.hw_device.empty())
@@ -1058,16 +1175,15 @@ int main(int argc, char *argv[])
         {
             use_dmabuf = true;
         } else if (force_no_dmabuf) {
-            std::cout << "Disabling DMA-BUF as requested on command line" << std::endl;
+            std::cerr << "Disabling DMA-BUF as requested on command line" << std::endl;
         } else {
-            std::cout << "compositor running on different device, disabling DMA-BUF" << std::endl;
+            std::cerr << "compositor running on different device, disabling DMA-BUF" << std::endl;
         }
 
-        // region with dmabuf not implemented in wlroots
-        if (selected_region.is_selected())
+        // region with dmabuf needs wlroots >= 0.17
+        if (use_dmabuf && selected_region.is_selected())
         {
-            use_dmabuf = false;
-            std::cout << "region capture not supported with DMA-BUF" << std::endl;
+            std::cerr << "region capture may not work with older wlroots, try --no-dmabuf if it fails" << std::endl;
         }
 
         if (params.video_filter == "null")
@@ -1081,7 +1197,7 @@ int main(int argc, char *argv[])
 
         if (use_dmabuf)
         {
-            std::cout << "enabled DMA-BUF capture, device " << params.hw_device.c_str() << std::endl;
+            std::cerr << "enabled DMA-BUF capture, device " << params.hw_device.c_str() << std::endl;
 
             drm_fd = open(params.hw_device.c_str(), O_RDWR);
             if (drm_fd < 0)
@@ -1162,9 +1278,6 @@ int main(int argc, char *argv[])
 
     printf("selected region %d,%d %dx%d\n", selected_region.x, selected_region.y, selected_region.width, selected_region.height);
 
-    timespec first_frame;
-    first_frame.tv_sec = -1;
-
     bool spawned_thread = false;
     std::thread writer_thread;
 
@@ -1203,12 +1316,7 @@ int main(int argc, char *argv[])
             spawned_thread = true;
         }
 
-        if (first_frame.tv_sec == -1)
-            first_frame = buffer.presented;
-
-        buffer.base_usec = timespec_to_usec(buffer.presented)
-            - timespec_to_usec(first_frame);
-
+        buffer.base_usec = timespec_to_usec(buffer.presented);
         buffers.next_capture();
     }
 
